#!/usr/bin/perl -w

# evoce (the Evolutionary Collage Engine) v0.1
# Copyright 2007 Brent Emerson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

use File::Fetch;
use Lingua::Treebank::Const;
$Lingua::Treebank::Const::STRINGIFY = "words";
$Lingua::Treebank::Const::INDENT_CHAR = '';
$Lingua::Treebank::Const::CHILD_PROLOG = '';
$Lingua::Treebank::Const::CHILD_EPILOG = ' ';
$File::Fetch::WARN = 0;

# turn this on if you'd like some fairly verbose detail about what's happening
$debug = 0;

$generation = 0;
$populationscale = 0;

print "\nWelcome to evoce, the Evolutionary Collage Engine.\n";
while (1 eq 1) {
	print "\nYou're working on generation $generation.\nWould you like to:\n\n\t\t(r)eview past generations\n\t\t(c)reate the next generation\n\t\t(i)mport into the current generation\n\t\tsave sentences in the (n)otebook or a (f)ile\n\t\t(q)uit?\n\nevoce > ";
	chomp($selection=<STDIN>);
	if ($selection eq "c") {
		algorithm()
	} elsif ($selection eq "i") {
		import()
	} elsif ($selection eq "f") {
		export()
	} elsif ($selection eq "r") {
		review()
	} elsif ($selection eq "n") {
		notebook()
	} elsif ($selection eq "q") {
		die "\nThanks for playing with evoce!\n"
	};
}


sub import {

print "\nDo you want to import from:\n\n\t\ta local plain (t)ext file\n\t\ta plain text file available at an (i)nternet URL\n\t\ta previously saved evoce (s)entence file?\n\nevoce > ";
chomp($selection=<STDIN>);
if ($selection eq "t") {
	print "\nEnter the filename of the plain text file you'd like to import:\n\nevoce > ";
	chomp($textfile=<STDIN>);
	$parsedfile = parse($textfile);
} elsif ($selection eq "i") {
	print "\nEnter the URL pointing to the plain text file you'd like to import:\n\nevoce > ";
	chomp($url=<STDIN>);
	my $ff = File::Fetch->new(uri => $url);
	if (my $where = $ff->fetch(to => '/tmp')) {
		$parsedfile = parse($where);
	} else {
		print "Your file was not found.  Please check the URL and try again.\n\n";
		return;
	}
} elsif ($selection eq "s") {
	print "\nEnter the filename of the evoce sentence file you'd like to import:\n\nevoce > ";
	chomp($parsedfile=<STDIN>);
}

# open parser output, create sentence objects

print "\nLoading sentences into generation $generation...\n";
if ($debug) { print "\nAttempting to open $parsedfile for import.\n"; };

@import = ();
open (PARSEROUTPUT, "$parsedfile");
while (<PARSEROUTPUT>) {
	chomp;
	push @import, Lingua::Treebank::Const->new->from_penn_string($_);
}
close (PARSEROUTPUT);

# present them to the author for approval/thinning

$doneimporting = 0;
until ($doneimporting) {
	print "\nHere are the imported sentences:\n";
	for $i (0 .. $#import) {
		print $i, "\t", $import[$i]->stringify(), "\n";
	}

	print "\nShould any of them be removed?\nEnter the sentence number to remove, or hit Enter to proceed.\n\nevoce > ";
	chomp($selection=<STDIN>);
	if ($selection eq '') {
		$doneimporting = 1;
	} else {
		splice (@import, $selection, 1);
		print "\nRemoved sentence $selection\n";
	}
}

# push the author-approved import into the current generation index (+diversity)

push @{$index[$generation]}, @import;
push @diversity, @import;
$populationscale += $#import + 1;
if ($debug) { print "\nPopulation scale is now $populationscale.\n"; };

print "\nHere are your generation $generation sentences:\n\n";
for $i (0 .. $#{$index[$generation]}) {
	print $i, "\t", $index[$generation][$i]->stringify(), "\n";
}

}


sub parse {
	my ($inputfile) = @_;

my $fileseed = int(rand 1000000);
my $filename1 = "/tmp/evoce-$fileseed-1";
my $filename2 = "/tmp/evoce-$fileseed-2";
my $filename3 = "/tmp/evoce-$fileseed-3";

# parse the input file's raw text into an array of sentence strings

print "\nPlease wait while the natural language parser works on your text.\nThis may take a long time (1-2 seconds per sentence)...\n";
if ($debug) { print "\nAttempting to open $inputfile for parsing.\n"; };

$wholeinput = '';
open (RAWINPUT, "$inputfile");
while (<RAWINPUT>) {
	chomp;
	$wholeinput = join(' ', $wholeinput, $_);
}
close (RAWINPUT);
$wholeinput =~ s/\((.*?)\)//g;
chomp($wholeinput);
chop($wholeinput);
@arrayofrawsentences = split (/[\.!?;] +/, $wholeinput);

# write the sentence strings in parser input format

open (PARSERINPUT, ">$filename1");
while ($rawsentence = shift @arrayofrawsentences) {
	print PARSERINPUT "<s> " . $rawsentence . " </s>\n";
}
close (PARSERINPUT);

# call parser to parse sentence strings with Penn tree-bank style tags

system ("parser/PARSE/parseIt -T50 parser/DATA/EN/ $filename1 > $filename2");
system ("grep S1 $filename2 | sed 's/^(S1 //' | sed 's/)\$//' > $filename3");

# return the name of the file containing the parsed sentences

print "\nParsing is complete!\n";
return $filename3;

}


sub export {

print "\nDo you want to save:\n\n\t\tthe sentences in the (n)otebook\n\t\tthe (o)riginally imported generation 0 sentences\n\t\tthe (c)urrent generation of sentences\n\t\t(a)ll sentences?\n\nevoce > ";
chomp($sentences=<STDIN>);

print "\nDo you want to save your sentences in\n\n\t\ta plain (t)ext file (easier to read)\n\t\tan evoce (s)entence file (suitable for later import by evoce)?\n\nevoce > ";
chomp($filetype=<STDIN>);
if ($filetype eq "t") {
	$method = "stringify";
} else {
	$method = "as_penn_text";
}

print "\nEnter the file you'd like to save to:\n\nevoce > ";
chomp($filename=<STDIN>);

open (EXPORT, ">$filename");
if ($sentences eq "n") {
	for $i (0 .. $#notebook) {
		print EXPORT $notebook[$i]->$method, "\n";
	}
} elsif ($sentences eq "o") {
	for $i (0 .. $#{$index[0]}) {
		print EXPORT $index[0][$i]->$method, "\n";
	}
} elsif ($sentences eq "c") {
	for $i (0 .. $#{$index[$generation]}) {
		print EXPORT $index[$generation][$i]->$method, "\n";
	}
} elsif ($sentences eq "a") {
	for $i (0 .. $#index) {
		for $j (0 .. $#{$index[$i]}) {
			print EXPORT $index[$i][$j]->$method, "\n";
		}
	}
}
close (EXPORT);

}


sub algorithm {

# check to make sure we have anything to work on

unless ($#{$index[$generation]} > 0) {
	print "\nYou don't have any sentences to breed!  You need to import some sentences first.\n";
	return;
}

## assign fitnesses to current generation index and fill population
## cross over current population and create/show next generation index

print <<EOF;

I'm about to show you the sentences in generation $generation.
Please rank each sentence from 1-5:
(1=awful 2=poor 3=fair 4=good 5=transcendent r=repeat m=menu)

EOF

# loop through sentences in the current index.  for each:
# ask for user input on fitness level n (1-5)
# push n-1 copies of sentence into population[g]

for $i (0 .. $#{$index[$generation]}) {
	print $index[$generation][$i]->stringify(), "\n";
	print "evoce > ";
	chomp($fitness=<STDIN>);
	if ($fitness =~ /[234]/ ) {
		push @{$population[$generation]}, ($index[$generation][$i]) x ($fitness - 1);
	} elsif ($fitness eq "5") {
		push @{$population[$generation]}, ($index[$generation][$i]) x 4;
		push @diversity, $index[$generation][$i];
	} elsif ($fitness eq "m") {
		$population[$generation] = ();
		return;
	};
}

# push the diversity array (all imports & 5-judged sentences) onto population[g]

push @{$population[$generation]}, @diversity;

if ($debug) {
	print "\nHere's the generation $generation population:\n";
	for $i (0 .. $#{$population[$generation]}) {
		print $i, "\t", $population[$generation][$i]->stringify(), "\n";
	}
	sleep 5;
};

## mate sentences in population[g] to create index[g+1]

$children = 0;
while ($children <= $populationscale) {

# randomly pick one of population[g]'s sentences

	$r = int(rand (1 + $#{$population[$generation]}));
	if ($debug) { print "\nSelected first sentence id: ", $r, "\nString: ", $population[$generation][$r]->stringify(), "\nTree: ", $population[$generation][$r]->as_penn_text(), "\n"; };

# randomly pick one of this sentence's (non-S) substructures
# (flatten the sentence's tree by recursively walking it and pushing non-S
# substructures onto a list. then pick a random element of the flattened tree)

	my $action = sub {
		my ($self,$state) = @_;
		push (@{$state}, $self) unless $self->tag() eq 'S';
	};
	my $stop_crit = sub {'0' eq '1'};
	@substructs = ();
	$population[$generation][$r]->walk($action,$stop_crit,\@substructs);
	$s = int(rand(1+ $#substructs));
	if ($debug) { print "\nSelected substructure ", $s, " of ", $#substructs, ": ", $substructs[$s]->tag(), "\nString: ", $substructs[$s]->stringify(), "\nTree: ", $substructs[$s]->as_penn_text(), "\n"; };

# repeat until the first sentence/substructure successfully mates with a
# second sentence or we've tried too many times to find a mate

	$attempts = 0;
	$keeptrying = 1;
	while ($keeptrying and ($attempts <= $populationscale)) {

# randomly pick a second population[g] sentence

		$t = int(rand (1 + $#{$population[$generation]}));
		if ($debug) { print "\nSelected second sentence id: ", $t, "\nString: ", $population[$generation][$t]->stringify(), "\nTree: ", $population[$generation][$t]->as_penn_text(), "\n"; };

# does a substructure with the same tag exist in the second sentence?
# (make a copy & walk its tree, pushing matching substructures onto a list)

		my $workingcopy = Lingua::Treebank::Const->new->from_penn_string($population[$generation][$t]->as_penn_text());
		my $action = sub {
			my ($self,$state) = @_;
			return unless $self->tag() eq $substructs[$s]->tag();
			push @{$state}, $self;
		};
		@matches = ();
		$workingcopy->walk($action,$stop_crit,\@matches);

# if there are 1 or more matching substructures, randomly select one and
# replace it with the selected substructure from the first sentence.
# push the resulting new sentence onto index[g+1]

		if ($#matches >= 0) {
			$u = int(rand (1 + $#matches));
			if ($debug) { print "\nSelected match ", $u, " of ", $#matches, "\nString: ", $matches[$u]->stringify(), "\nTree: ", $matches[$u]->as_penn_text(), "\nParent is ", $matches[$u]->parent()->as_penn_text(), "\n"; };
			$matches[$u]->parent()->replace($matches[$u],$substructs[$s]);
			if ($debug) { print "\nResult of mating\nString: ", $workingcopy->stringify(), "\nTree: ", $workingcopy->as_penn_text(), "\n"; };
			push @{$index[$generation+1]}, $workingcopy;
			++$children;
			$keeptrying=0;
		} else {
			if ($debug) { print "\nNo matches\n"; };
		}
		++$attempts;
	}
}

# increment generation, print out the new index

++$generation;
print "\nHere are your generation $generation sentences:\n";
for $i (0 .. $#{$index[$generation]}) {
	print $i, "\t", $index[$generation][$i]->stringify(), "\n";
}

}


sub review {

for $i (0 .. $#index) {
	print "\nGENERATION $i\n";
	for $j (0 .. $#{$index[$i]}) {
		print $j, "\t", $index[$i][$j]->stringify(), "\n";
	}
	print "\n";
}

}


sub notebook {

while (1 eq 1) {
	print "\nWhich sentence from the current generation would you like to store in the notebook?  You can also\n\n\t\t(r)eview the notebook\n\t\treturn to the (m)ain menu.\n\nevoce > ";
	chomp($selection=<STDIN>);
	if ($selection eq "r") {
		print "\nNOTEBOOK\n";
		for $i (0 .. $#notebook) {
			print $i, "\t", $notebook[$i]->stringify(), "\n";
		}
		print "\n";
	} elsif ($selection eq "m") {
		return;
	} else {
		push @notebook, $index[$generation][$selection];
		print "\n\"$index[$generation][$selection]\" has been stored in the notebook.\n";
	};
}

}
